/* 
   v4l1_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h> 
#include <fcntl.h> 
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h> 
#include <errno.h>    
#include <unistd.h>   
#include <linux/videodev.h>
#include <pthread.h>

#include "NVrec.h"

struct vbuf_t {
	unsigned char **buf;	// pointer to memory of each buffer
	char full;		// buffer SYNC'ed
	ntime_t ts;		// buffer SYNC'ed
	unsigned char *mem;	// memory if we need to bounce it
};

struct v4l1_core_t {
	int fd;			// file descriptor
	
	int width;		// image width we got
	int height;		// image height we got
	ntime_t frame_period;	// time for each frame
	int planes;		// number of planes/rows per image
	int framesize;		// size of frames
	
 	struct video_mbuf vmb;	// mmap buffers
	int nbufs;		// number of buffers received
	unsigned char *mmbuf;	// memory map buffer
	struct vbuf_t *vbufs;	// the individual buffers

	int last_index;		// index of last frame dequeued
	int last_seq;		// last sequence number
	ntime_t last_time;	// timestamp of last frame
	int new_index;		// index of frame dequeued
	int new_seq;		// new sequence number
	ntime_t new_time;	// timestamp of new frame
	
	ntime_t start_time;	// timestamp of first frame
	
	struct video_mmap mm;	// temp buffer
	
	int exit;		// stop SYNC thread
	int done;		// SYNC thread terminated

	int sync;		// sync end
	int queue;		// queue end
	int read;		// read point
	
	int copyread;		// will copy buffers
};

void * v4l1copy_core_thread(void * t)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)t;
	int i;
	int frame = 0;
	ntime_t ttime;
	struct sched_param sp;
	
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		perror("v4l1 core thread - could not get FIFO priority");
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			perror("v4l1 core thread - could not get RR priority");
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				perror("v4l1 core thread - could not get any priority");
			}
		}
	}
	
	while(!vct->exit) {
		i = -1;
		while(i < 0) {
			i = ioctl(vct->fd, VIDIOCSYNC, &frame);
			if(i < 0 && errno == EINTR) continue;
			if(i < 0) {
				perror("v4l1 core next - error dequeueing buffer (trying to continue)");
				vct->mm.frame = frame;
				ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm);
				frame++;
				if(frame >= vct->vmb.frames)
					frame = 0;
			}
		}
		ttime = getntime();
		if((vct->vbufs + vct->sync)->full == 1) { // Nothing left to sync!
			fprintf(stderr, "v4l1 core thread - no place to put new data - discarding %d\n", vct->exit);
			vct->mm.frame = frame;
			if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
				perror("v4l1 core init - could not queue new buffer");
			frame++;
			if(frame >= vct->vmb.frames)
				frame = 0;
			usleep(vct->frame_period / 1000);
			continue;
		}
		memcpy((vct->vbufs + vct->sync)->mem, vct->mmbuf + vct->vmb.offsets[frame], vct->framesize);
		vct->mm.frame = frame;
		if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
			perror("v4l1 core init - could not queue new buffer");
		frame++;
		if(frame >= vct->vmb.frames)
			frame = 0;
		(vct->vbufs + vct->sync)->ts = ttime;
		(vct->vbufs + vct->sync)->full = 1;
		vct->sync++;
		if(vct->sync >= vct->nbufs)
			vct->sync = 0;
	}
	fprintf(stderr, "v4l1 core thread - exiting!\n");
	vct->done = 1;
	return NULL;
}

void * v4l1_core_thread(void * t)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)t;
	int i;
	struct sched_param sp;
	
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		perror("v4l1 core thread - could not get FIFO priority");
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			perror("v4l1 core thread - could not get RR priority");
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				perror("v4l1 core thread - could not get any priority");
			}
		}
	}
	while(!vct->exit) {
		if((vct->vbufs + vct->sync)->full == 1) { // Nothing left to sync!
			 fprintf(stderr, "v4l1 core thread - all buffers SYNCed - discarding data %d\n", vct->exit);
			 usleep(vct->frame_period / 1000);
			 continue;
		}
		i = -1;
		while(i < 0) {
			i = ioctl(vct->fd, VIDIOCSYNC, &vct->sync);
			if(i < 0 && errno == EINTR) continue;
			if(i < 0) {
				perror("v4l1 core next - error dequeueing buffer (trying to continue)");
				i = 0;
			}
		}
		(vct->vbufs + vct->sync)->ts = getntime();
		(vct->vbufs + vct->sync)->full = 1;
		vct->sync++;
		if(vct->sync >= vct->nbufs)
			vct->sync = 0;
	}
	fprintf(stderr, "v4l1 core thread - exiting!\n");
	vct->done = 1;
	return NULL;
}

void v4l1_core_freevbufs(volatile struct v4l1_core_t * vct)
{
	if(!vct->vbufs)
		return;
	while(vct->nbufs > 0) {
		vct->nbufs--;
		if((vct->vbufs + vct->nbufs)->buf)
			free((vct->vbufs + vct->nbufs)->buf);
		if((vct->vbufs + vct->nbufs)->mem)
			free((vct->vbufs + vct->nbufs)->mem);
	}
	free(vct->vbufs);
}

void v4l1_core_usage()
{
	fprintf(stderr, "V4L1 capture core usage:\n");
	fprintf(stderr, "\t-v <device name> capture device name  [/dev/video]\n");
	fprintf(stderr, "\t-w <width> picture width              [384]\n");
	fprintf(stderr, "\t-h <height> picture height            [288]\n");
	fprintf(stderr, "\t-norm <PAL|NTSC|SECAM> set input norm [unchanged]\n");
	fprintf(stderr, "\t-input <input name> set input         [unchanged]\n");
	fprintf(stderr, "\t-frequency <MHz> set input frequency  [unchanged]\n");
}

struct cap_core_t * v4l1_core_init(char ** args, unsigned int pfmt)
{
	struct cap_core_t * cct;
	struct v4l1_core_t * vct;
 	int i, j;
 	pthread_t vthread;
 	struct video_window vwin;
 	struct video_audio vaud;
	char * fdev = NULL;
	int N = -1;
	char * norm = NULL;
	char * inp = NULL;
	double freq = -1;
	NV_ARGS;

	cct = malloc(sizeof(struct cap_core_t));
	if(!cct) {
		perror("v4l1 core init - could no malloc new cap core");
		return NULL;
	}
	vct = malloc(sizeof(struct v4l1_core_t));
	if(!vct) {
		perror("v4l1 core init - could no malloc new v4l2 core");
		free(cct);
		return NULL;
	}
	memcpy(cct, &v4l1_core, sizeof(struct cap_core_t));
	cct->priv = (void *)vct;
	memset(vct, 0, sizeof(struct v4l1_core_t));
 
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vct->width = atoi(NV_THIS_ARG);
			if(vct->width < 16)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vct->height = atoi(NV_THIS_ARG);
			if(vct->height < 16)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-v") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fdev = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-N") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			N = atoi(NV_THIS_ARG);
			if(N < 4)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-norm") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			norm = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-input") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			inp = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-frequency") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			freq = atof(NV_THIS_ARG);
			if(freq < 0)
				goto usage;
		}
	}
	fprintf(stderr, "v4l1 core init - Size: %dx%d from %s\n", vct->width, vct->height, fdev);

	vct->fd = open(fdev, O_RDWR);
	if(vct->fd<=0) {
		perror("v4l1 core init - open capture device");
		goto error0;
	}

	if(inp) {
		struct video_channel chan;
		int i = 0;
		
		do {
			chan.channel = i;
			if(ioctl(vct->fd, VIDIOCGCHAN, &chan)) {
				fprintf(stderr, "v4l1 core init - unknown input '%s'\n", inp);
				goto error1;
			}
			if(strcmp(chan.name, inp) == 0)
				break;
			i++;
		}  while (i > 0);
		if(norm) {
			if(strcmp(norm, "PAL") == 0) {
				chan.norm = VIDEO_MODE_PAL;
			} else if(strcmp(norm, "NTSC") == 0) { 
				chan.norm = VIDEO_MODE_NTSC;
			} else if(strcmp(norm, "SECAM") == 0) { 
				chan.norm = VIDEO_MODE_SECAM;
			} else {
				fprintf(stderr, "v4l1 core init - unknown norm '%s'\n", norm);
				goto error1; 
			}
		}
		if(ioctl(vct->fd, VIDIOCSCHAN, &chan)) {
			perror("v4l1 core init - could not set input");
			goto error1;
		}
		switch(chan.norm) {
			case VIDEO_MODE_SECAM:
			case VIDEO_MODE_PAL:
				vct->frame_period = 40000000ULL;
				break;
			case VIDEO_MODE_NTSC:
				vct->frame_period = 33366667ULL;
				break;
		}
	} else if(norm) {
		fprintf(stderr, "Sorry, v4l1 sucks.  If you want to set the norm, you must set the input too.\n");
		goto error1;
	} else {
		vct->frame_period = 40000000ULL; /* Default to PAL */
	}
	
	if(freq > 0) {
		int i = (int)(freq * 16.0);
		if(ioctl(vct->fd, VIDIOCSFREQ, &i)) {
			perror("v4l1 core init - could not set frequency");
			goto error1;
		}
	}
 
 	vwin.x = 0;
 	vwin.y = 0;
 	vwin.flags = 0;
 	vwin.width = vct->width;
 	vwin.height = vct->height;
 	vwin.clips = NULL;
 	vwin.clipcount = 0;
	if (ioctl(vct->fd, VIDIOCSWIN, &vwin)) {
		perror("v4l1 core init - set window");
		goto error1;
	}
	vct->mm.width = vct->width;
	vct->mm.height = vct->height;
	switch (pfmt) {
		case NV_RGB24:
			vct->mm.format = VIDEO_PALETTE_RGB24;
			vct->planes = vct->height;
			vct->framesize = vct->width * vct->height * 3;
			break;
		case NV_YUV420:
			vct->mm.format = VIDEO_PALETTE_YUV420P;
			vct->planes = 3;
			vct->framesize = (vct->width * vct->height * 3) / 2;
			break;
		case NV_YUV422:
			vct->mm.format = VIDEO_PALETTE_YUV422;
			vct->planes = vct->height;
			vct->framesize = vct->width * vct->height * 2;
			break;
		default:
			fprintf(stderr, "v4l1 core init - unsupported video format\n");
			goto error1;
	}
	if (ioctl(vct->fd, VIDIOCGMBUF, &vct->vmb) != 0) {
		perror("v4l1 core init - could not request buffers");
		goto error1;
	}
	if (vct->vmb.frames == 0) {
		fprintf(stderr, "v4l1 core init - could not get any buffers\n");
		goto error1;
	}
	if (vct->vmb.frames < N) {
		fprintf(stderr, "v4l1 core init - could not get enough buffers - using copy mode\n");
		i = N;
		vct->copyread = 1;
	} else {
		i = vct->vmb.frames;
	}
	vct->mmbuf = (unsigned char *)mmap(NULL, vct->vmb.size, PROT_READ|PROT_WRITE, MAP_SHARED, vct->fd, 0);
	if((int)(vct->mmbuf) == -1) {
		perror("v4l1 core init - mmap failed");
		goto error1;
	}

	vct->vbufs = (struct vbuf_t *)malloc(sizeof(struct vbuf_t) * i);
	if(!vct->vbufs) {
		fprintf(stderr, "v4l1 core init - could not allocate memory (vbufs)");
		goto error1;
	}
	memset(vct->vbufs, 0, sizeof(struct vbuf_t) * i);
	while(vct->nbufs < i) {
		(vct->vbufs + vct->nbufs)->buf = (unsigned char **)malloc(sizeof(unsigned char *) * vct->planes);
		if(!(vct->vbufs + vct->nbufs)->buf) {
			perror("v4l1 core init - could not allocate buffer pointers");
			goto error1;
		}
		if(vct->copyread) {
			(vct->vbufs + vct->nbufs)->mem = (unsigned char *)malloc(vct->framesize);
			if(!(vct->vbufs + vct->nbufs)->mem) {
				perror("v4l1 core init - could not allocate buffer memory");
				goto error1;
			}
			(vct->vbufs + vct->nbufs)->buf[0] = (vct->vbufs + vct->nbufs)->mem;
		} else {
			(vct->vbufs + vct->nbufs)->buf[0] = vct->mmbuf + vct->vmb.offsets[vct->nbufs];
		}
		switch(pfmt) {
			case NV_YUV420:
				(vct->vbufs + vct->nbufs)->buf[1] =
					(vct->vbufs + vct->nbufs)->buf[0] +
					vct->width * vct->height;
				(vct->vbufs + vct->nbufs)->buf[2] =
					(vct->vbufs + vct->nbufs)->buf[1] +
					(vct->width * vct->height) / 4;
				break;
			case NV_RGB24:
				for(j = 1; j < vct->height; j++) {
					(vct->vbufs + vct->nbufs)->buf[j] =
						(vct->vbufs + vct->nbufs)->buf[j - 1]
						+ vct->width * 3;
				}
				break;
			case NV_YUV422:
				for(j = 1; j < vct->height; j++) {
					(vct->vbufs + vct->nbufs)->buf[j] =
						(vct->vbufs + vct->nbufs)->buf[j - 1]
						+ vct->width * 2;
				}
				break;
		}
		vct->nbufs++;
	}
	fprintf(stderr, "v4l1 core init - got %i buffers\n", vct->vmb.frames);
	if(vct->copyread)
		fprintf(stderr, "v4l1 core init - emulating %i buffers\n", vct->nbufs);

	if(ioctl(vct->fd, VIDIOCGAUDIO, &vaud)!=0)
		perror("v4l1 core init - could not get mute status");
	vaud.flags &= ~VIDEO_AUDIO_MUTE;
	if(ioctl(vct->fd, VIDIOCSAUDIO, &vaud)!=0)
		perror("v4l1 core init - could not set mute status");

	for(i = 0; i < vct->vmb.frames; i++) {
		vct->mm.frame = i;
		if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0) {
			perror("v4l1 core init - could not queue new buffer");
			goto error1;
		}
	}

	vct->last_index = -1;
	vct->last_seq = -1;
	vct->new_index = -1;
	vct->new_seq = -1;
	
	vct->exit = 0;
	vct->done = 0;
	vct->sync = 0;
	vct->queue = 0; 
	vct->read = 0;
	
	if(vct->copyread) {
		if(pthread_create(&vthread, NULL, v4l1copy_core_thread, (void *)vct) != 0) {
			perror("v4l1 core init - could not start v4l1 thread");
			goto error1;
		}
	} else {
		if(pthread_create(&vthread, NULL, v4l1_core_thread, (void *)vct) != 0) {
			perror("v4l1 core init - could not start v4l1 thread");
			goto error1;
		}	
	}
	return cct;
error1:
	v4l1_core_freevbufs(vct);
	close(vct->fd);
error0:
	free(cct);
	free(vct);
	return NULL;
usage:
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(cct);
	free(vct);
	return NULL;
}

void v4l1_core_close(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;
 	struct video_audio vaud;

	if(!vct)
		return;
	vct->exit = 1;
	while(!vct->done)
		usleep(100000);
	fprintf(stderr, "v4l1 core close - SYNC thread terminated.\n");

	if(ioctl(vct->fd, VIDIOCGAUDIO, &vaud)!=0)
		perror("v4l1 core init - could not get mute status");
	vaud.flags |= VIDEO_AUDIO_MUTE;
	if(ioctl(vct->fd, VIDIOCSAUDIO, &vaud)!=0)
		perror("v4l1 core init - could not set mute status");

	v4l1_core_freevbufs(vct);
	close(vct->fd);
	free((void *)vct);
	free(cct);
	return;
}

unsigned char ** v4l1_core_next(struct cap_core_t * cct, ntime_t * ftime)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;

	if (!vct)
		return NULL;
	if (vct->new_index < 0) { // FIXME get out of this loop on INT
		while((vct->vbufs + vct->read)->full == 0) {
			usleep(10000); // wait for new frame
		}
		vct->new_index = vct->read;
		vct->new_time = (vct->vbufs + vct->new_index)->ts;
		if(vct->last_index < 0) {
			vct->start_time = vct->new_time;
			vct->new_seq = 0;
		} else {
			vct->new_seq =
				( (vct->new_time - vct->start_time)
				+ (vct->frame_period >> 1) ) / 
				vct->frame_period;
		}
		vct->read++;
		if(vct->read >= vct->nbufs)
			vct->read = 0;
	}
	if(vct->last_index < 0) {
		/* First frame */
		if (vct->new_index < 0) return NULL;
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		vct->new_index = -1;
		*ftime = vct->last_time;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	} else if(vct->new_seq == (vct->last_seq + 1)) {
		/* New frame, in sequence */
		if(!vct->copyread) {
			vct->mm.frame = vct->last_index;
			if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
				perror("v4l1 core next - could not requeue old buffer");
		}
		(vct->vbufs + vct->last_index)->full = 0;
		vct->last_index = vct->new_index;
		vct->last_seq = vct->new_seq;
		vct->last_time = vct->new_time;
		*ftime = vct->last_time;
		vct->new_index = -1;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	} else if(vct->new_seq <= vct->last_seq)
	{
		/* New frame out of sequence! */
		/* FIXME - we can get some strange frame ordering here! */
		fprintf(stderr, "v4l1 core next - new frame out of sequence - ignoring\n");
		if(vct->new_index >= 0) {
			if(!vct->copyread) {
				vct->mm.frame = vct->last_index;
				if(ioctl(vct->fd, VIDIOCMCAPTURE, &vct->mm)!=0)
					perror("v4l1 core next - could not requeue old buffer");
			}
			(vct->vbufs + vct->last_index)->full = 0;
			vct->last_index = vct->new_index;
			vct->new_index = -1;
		}
		return NULL;
	} else {
		/* New frame - lost something between though */
		vct->last_seq++;
		vct->last_time += vct->frame_period;
		*ftime = vct->last_time;
		return &((vct->vbufs + vct->last_index)->buf[0]);
	}
	return NULL;
}

double v4l1_core_status(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;
	double qd;

	if(!vct)
		return -1;
	if(vct->new_index >= 0) {
		qd = vct->new_time;
	} else if(vct->last_index >= 0) {
		qd = vct->last_time;
	} else {
		return 0;
	}
	qd = (double)getntime() - qd;
	qd /= (double)vct->nbufs * (double)vct->frame_period;
	return qd * 100.0;
}

inline ntime_t v4l1_core_ftime(struct cap_core_t * cct)
{
	volatile struct v4l1_core_t * vct = (struct v4l1_core_t *)cct->priv;

	if(!vct)
		return -1;
	return vct->frame_period;
}

struct cap_core_t v4l1_core = {
	cap_core_init: v4l1_core_init,
	cap_core_close: v4l1_core_close,
	cap_core_next: v4l1_core_next,
	cap_core_status: v4l1_core_status,
	cap_core_ftime: v4l1_core_ftime,
	cap_core_usage: v4l1_core_usage,
	priv: NULL
};
